/* -*- Mode: C; tab-width:4 -*- */
/* ex: set ts=4 shiftwidth=4 softtabstop=4 cindent: */
/**
 * @brief Example module for SOS
 * This module shows some of concepts in SOS
 */

/**
 * Module needs to include <module.h>
 */
//#undef _MODULE_

#include <sys_module.h>

//#define LED_DEBUG
#include <led_dbg.h>

#include "unary.h"

#include <MassWare/MassWare/massware.h>

#define BLINK_TIMER_INTERVAL	512L
#define BLINK_YELLOW_TID        0

/**
 * Module can define its own state
 */
typedef struct {
  uint8_t pid;
  uint8_t state;
  MassOutputInterf * outfs;
  uint8_t outf_cnt;
  uint8_t lswt_scale_level; // scale level is used to calcuate the DSC value length
  uint16_t data_length;
  data_type *data_set;
  uint16_t compressed_data_length;
  uint8_t *compressed_data_set;
  uint16_t pkt_size; // the packet size and parcket are used to record the processed data and sent them
  void * pkt;        // after they are process and red LED blinked
} app_state_t;

static int8_t unary_msg_handler(void *start, Message *e);

/**
 * This is the only global variable one can have.
 */
static const mod_header_t mod_header SOS_MODULE_HEADER = {
	.mod_id         = UNARY_CODING_MOD_ID,
	.state_size     = sizeof(app_state_t),
	.num_sub_func   = 0,
	.num_prov_func  = 0,
	.platform_type  = HW_TYPE /* or PLATFORM_ANY */,
	.processor_type = MCU_TYPE,
	.code_id        = ehtons(UNARY_CODING_MOD_ID),
	.module_handler = unary_msg_handler,
};


static int8_t unary_msg_handler(void *state, Message *msg)
{
	/**
	* The module is passed in a void* that contains its state.  For easy
	* reference it is handy to typecast this variable to be of the
	* applications state type.  Note that since we are running as a module,
	* this state is not accessible in the form of a global or static
	* variable.
	*/
	app_state_t *s = (app_state_t*)state;

	/**
	* Switch to the correct message handler
	*/
	switch (msg->type){
		/**
		* MSG_INIT is used to initialize module state the first time the
		* module is used.  Many modules set timers at this point, so that
		* they will continue to receive periodic (or one shot) timer events.
		*/
  case MSG_INIT:
	  {
		  s->pid = msg->did;
		  s->state = 0;
		  s->outfs = NULL;
		  s->outf_cnt = 0;
		  s->data_length = 0;
		  s->data_set = NULL;

		  sys_led(LED_YELLOW_OFF);
		  /**
		  * The timer API takes the following parameters:
		  * - Timer ID (used to distinguish multiple timers of different
		  *   ..types on the same host)
		  * - Timer delay in
		  * - Type of timer
		  */
		  //	  DEBUG("Green Blink Start\n");
		  //	  sys_timer_start(BLINK_GREEN_TID, BLINK_TIMER_INTERVAL, TIMER_REPEAT);
		  break;
	  }

  case MSG_FORWARD_INPUT:
	  {
		  sys_led(LED_YELLOW_OFF);

		  // input are data_type type from lswt module, we keep it as data_type

		  // the first uint16 (2 bytes) is the data length, the thrid byte is the scale number
		  // the rest are the data payload,
		  s->data_length = ehtons(*(uint16_t *)(msg->data));
		  s->lswt_scale_level = *((uint8_t *)(msg->data) + 2);
/*		  data_type * input = (data_type *)((uint8_t *)(msg->data) + 3);

		  // the length of the large value data set;
		  uint16_t dsc_length = s->data_length>>s->lswt_scale_level;
		  
		  // number of bits to encode the position difference of two unzero data in the unary coding algorithm
  		  uint8_t POS_BIT = 0;
		  while ((1<<POS_BIT) < s->data_length) POS_BIT++;
		  if (POS_BIT > 9) POS_BIT = 9;

		  s->compressed_data_set = (uint8_t *)sys_malloc(s->data_length);
		  if (NULL == s->compressed_data_set) break;

		  s->compressed_data_length = 0;

		  // get the bit number used to code the data set (0 ~ data_length>>scale_level)
		  data_type amin, amax;
		  amin = amax = input[0];
		  uint16_t i;
		  for (i = 1; i < dsc_length; i++) {
			  if (input[i] < amin) 
				  amin = input[i];
			  else if (input[i] > amax)
				  amax = input[i];
		  }
		  uint8_t nbase = 0;
		  while((1<<nbase)<(amax-amin)) nbase++;

		  // DSC compress the data (0 ~ data_length>>scale_level)
		  uint16_t codelen;
		  uint8_t bit_length, bit_value;
		  data_type data_value;

		  for (i = 0, codelen = 0, bit_length = 0, bit_value = 0; i<dsc_length; i++) {
			  data_value = input[i]-amin; 
			  if (bit_length >= (8-nbase)){
				  s->compressed_data_set[s->compressed_data_length++] = (data_value >> (nbase+bit_length-8)) | bit_value;
				  bit_value = data_value << (16-nbase-bit_length);
				  bit_length = nbase+bit_length-8;
			  }
			  else {
				  bit_value = bit_value | (data_value<<(8-nbase-bit_length));
				  bit_length += nbase;
			  }
		  }

		  // Unary compress the data (data_length>>scale_level ~ data_length)
		  uint16_t pos = dsc_length; // record the last encoded data position
		  for (i = dsc_length; i < s->data_length; i++) {
			  if(input[i] != 0){
				  //unary coding the data if it is not equal to 0
				  data_type temp;
				  if(input[i]<0) 
					  temp = (0 - input[i]) * 2 - 1;
				  else
					  temp = input[i] * 2;

				  data_value = (2<<temp)-2;
				  bit_length = temp + 1 +bit_length;

				  if (bit_length >= 8){
					  s->compressed_data_set[s->compressed_data_length++] = (data_value>>(bit_length-8))|bit_value;
					  bit_length -= 8;
					  if(bit_length >= 8){
						  s->compressed_data_set[s->compressed_data_length++] = (data_value>>(bit_length-8))&255;
						  bit_length -= 8;
					  }
					  bit_value = (data_value<<(8-bit_length)) & 255;
				  }
				  else {//it must be the data
					  bit_value = bit_value|(data_value<<(8-bit_length));
				  }

				  // unary coding the data position
				  data_value = i-pos;
				  bit_length = POS_BIT + bit_length;

				  if (bit_length >= 8){
					  s->compressed_data_set[s->compressed_data_length++] = (data_value>>(bit_length-8))|bit_value;
					  bit_length -= 8;
					  if(bit_length >= 8){
						  s->compressed_data_set[s->compressed_data_length++] = (data_value>>(bit_length-8))&255;
						  bit_length -= 8;
					  }
					  bit_value = (data_value<<(8-bit_length)) & 255;
				  }
				  else {
					  bit_value = bit_value|(data_value<<(8-bit_length));
				  }

				  pos = i;
			  }
		  }

		   //save the last byte which is filled out
		  if(bit_length>0) s->compressed_data_set[s->compressed_data_length++] = bit_value;

		  // get the packet size that should be sent out
		  // first two bytes for data length, the third for lswt scale number, 
		  // the fourth for the base number, the fifth and sixth for the minimum data value.
		  s->pkt_size = 6 + s->compressed_data_length;
		  s->pkt = sys_malloc(s->pkt_size); 
		  if (NULL == s->pkt) break;

		  // the first two bytes are data length, the third for lswt scale number, 
		  *(uint16_t*)(s->pkt) = s->data_length;
		  *((uint8_t*)(s->pkt) + 2) = s->lswt_scale_level;

		  // the fourth for dsc correlation number
		  *((uint8_t*)(s->pkt) + 3) = nbase;

		  // the fifth and sixth are for the minimum data value.
		  *(uint16_t*)((uint8_t*)(s->pkt) + 4) = amin;

		  // the rest are the compressed data payload
		  uint8_t * data_out = (uint8_t*)(s->pkt) + 6;

		  for (i = 0; i<s->compressed_data_length; i++) 
			  data_out[i] = s->compressed_data_set[i];

		  sys_free(s->compressed_data_set);
*/
		  s->pkt_size = 1;
		  s->pkt = sys_malloc(s->pkt_size); 
		  DEBUG("Yellow Blink Start\n");
		  sys_led(LED_YELLOW_TOGGLE);
		  sys_timer_start(BLINK_YELLOW_TID, BLINK_TIMER_INTERVAL, TIMER_ONE_SHOT);
		  break;
	  }
	  
	  /**
	  * set the output component id (mod_id) and the interface of the component (msg_type)
	  * so that masslets can be connected dynamically, used for arch reconfiguration
	  */    

  case MSG_FORWARD_OUTPUT:
	  {
		  // PostDestMsg *pdmsg = (PostDestMsg*)sys_msg_take_data(msg);
		  PostDestMsg *pdmsg = (PostDestMsg*)(msg->data);
		  MassOutputInterf * outf = search_interface(s->outfs, MSG_FORWARD_OUTPUT);
		  if (outf == NULL) {
			  outf = (MassOutputInterf*) sys_malloc(sizeof(MassOutputInterf));
			  outf->out_msg_type = MSG_FORWARD_OUTPUT;
			  outf->dst_mod_id = pdmsg->dst_mod_id;
			  outf->dst_msg_type = pdmsg->msg_type;
			  outf->next = s->outfs;
			  s->outfs = outf;
			  s->outf_cnt ++;
		  }
		  else {
			  outf->dst_mod_id = pdmsg->dst_mod_id;
			  outf->dst_msg_type = pdmsg->msg_type;
		  }
		  break;
	  }


  case MSG_INVERSE_INPUT:
	  break;
  case MSG_INVERSE_OUTPUT:
	  break;
	  
	  /**
	  * All timers addressed to this PID, regardless of the timer ID, are of
	  * type MSG_TIMER_TIMEOUT and handled by this handler.  Timers with
	  * different timer IDs can be further distinguished by testing for the
	  * type, as demonstrated in the relay module.
	  */
  case MSG_TIMER_TIMEOUT:
	  {
		  sys_led(LED_YELLOW_TOGGLE);
		  
		  // send the data set out, we do not need to release the data set because
		  // the send command will release the dataset
		  
		  MassOutputInterf * outf = search_interface(s->outfs, MSG_FORWARD_OUTPUT);
		  if (outf != NULL)
			  sys_post(outf->dst_mod_id, outf->dst_msg_type, s->pkt_size, s->pkt, SOS_MSG_RELEASE);
		  return SOS_OK;
	  }
	  
	  /**
	  * MSG_FINAL is used to shut modules down.  Modules should release all
	  * resources at this time and take care of any final protocol
	  * shutdown.
	  */
  case MSG_FINAL:
	  {
		  // release memory
		  release_interface(s->outfs);
		  s->outfs = NULL;
		  s->outf_cnt = 0;

		  sys_led(LED_YELLOW_OFF);
		  DEBUG("Yellow Blink Stop\n");
		  break;
	  }

	  /**
	  * The default handler is used to catch any messages that the module
	  * does no know how to handle.
	  */
  default:
	  return -EINVAL;
  }

  /**
  * Return SOS_OK for those handlers that have successfully been handled.
  */
  return SOS_OK;
}

#ifndef _MODULE_
mod_header_ptr unary_get_header()
{
	return sos_get_header_address(mod_header);
}
#endif


